from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat
)
from typing import (
    Dict, 
    Union,
    Callable
)


PrimaryType = Union[str, int, float, bytes, bool]


class Env(BaseModel):
    MERCURY_SYNC_TCP_CONNECT_RETRIES: StrictInt=3
    MERCURY_SYNC_BOOT_WAIT: StrictStr='3s'
    MERCURY_SYNC_MAX_TIME_IDLE: StrictStr='10s'
    MERCURY_SYNC_IDLE_REBOOT_TIMEOUT: StrictStr='10s'
    MERCURY_SYNC_MIN_SUSPECT_NODES_THRESHOLD=3
    MERCURY_SYNC_MIN_SUSPECT_TIMEOUT_MULTIPLIER: StrictInt=2
    MERCURY_SYNC_MAX_SUSPECT_TIMEOUT_MULTIPLIER: StrictInt=7
    MERCURY_SYNC_INITIAL_NODES_COUNT: StrictInt=3
    MERCURY_SYNC_HEALTH_CHECK_TIMEOUT: StrictStr='0.2s'
    MERCURY_SYNC_REGISTRATION_TIMEOUT: StrictStr='1m'
    MERCURY_SYNC_HEALTH_POLL_INTERVAL: StrictFloat='0.2s'
    MERCURY_SYNC_INDIRECT_CHECK_NODES: StrictInt=3
    MERCURY_SYNC_CLEANUP_INTERVAL: StrictStr='10s'
    MERCURY_SYNC_MAX_CONCURRENCY: StrictInt=2048
    MERCURY_SYNC_AUTH_SECRET: StrictStr

    @classmethod
    def types_map(self) -> Dict[str, Callable[[str], PrimaryType]]:
        return {
            'MERCURY_SYNC_TCP_CONNECT_RETRIES': int,
            'MERCURY_SYNC_MAX_TIME_IDLE': str,
            'MERCURY_SYNC_IDLE_REBOOT_TIMEOUT': str,
            'MERCURY_SYNC_MIN_SUSPECT_NODES_THRESHOLD': int,
            'MERCURY_SYNC_MIN_SUSPECT_TIMEOUT_MULTIPLIER': int,
            'MERCURY_SYNC_MAX_SUSPECT_TIMEOUT_MULTIPLIER': int,
            'MERCURY_SYNC_INITIAL_NODES_COUNT': int,
            'MERCURY_SYNC_BOOT_WAIT': str,
            'MERCURY_SYNC_REGISTRATION_TIMEOUT': str,
            'MERCURY_SYNC_HEALTH_POLL_INTERVAL': str,
            'MERCURY_SYNC_INDIRECT_CHECK_NODES': int,
            'MERCURY_SYNC_CLEANUP_INTERVAL': str,
            'MERCURY_SYNC_MAX_CONCURRENCY': int,
            'MERCURY_SYNC_AUTH_SECRET': str
        }


import asyncio
from collections import deque
from typing import Callable, Tuple, Deque


class MercurySyncUDPProtocol(asyncio.DatagramProtocol):
    def __init__(
        self, 
        callback: Callable[
            [
                bytes,
                Tuple[str, int]
            ],
            bytes
        ]
    ):
        super().__init__()
        self.callback = callback

    def connection_made(self, transport) -> str:
        self.transport = transport

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        # Here is where you would push message to whatever methods/classes you want.
        # data: Message = pickle.loads(lzma.decompress(unpacked))
        self.callback(
            data,
            addr
        )


from pydantic import (
    StrictStr,
    StrictInt
)
from typing import Literal, Optional, List
from .message import Message


HealthStatus = Literal[
    "initializing",
    "waiting",
    "healthy", 
    "suspect", 
    "degraded", 
    "failed"
]

class HealthCheck(Message):
    target_host: Optional[StrictStr]
    target_port: Optional[StrictInt]
    target_status: Optional[StrictStr]
    source_host: StrictStr
    source_port: StrictInt
    source_status: Optional[StrictStr]
    status: HealthStatus


from time import time
from typing import Optional
from .constants import (
    MAX_SEQ
)
from .snowflake import Snowflake


class SnowflakeGenerator:
    def __init__(
        self, 
        instance: int, 
        *, 
        seq: int = 0, 
        timestamp: Optional[int] = None
    ):

        current = int(time() * 1000)


        timestamp = timestamp or current

        self._ts = timestamp

        self._inf = instance << 12
        self._seq = seq

    @classmethod
    def from_snowflake(cls, sf: Snowflake) -> 'SnowflakeGenerator':
        return cls(sf.instance, seq=sf.seq, epoch=sf.epoch, timestamp=sf.timestamp)

    def __iter__(self):
        return self

    def generate(self) -> Optional[int]:

        current = int(time() * 1000)

        if self._ts == current:

            if self._seq == MAX_SEQ:
                return None
            
            self._seq += 1

        elif self._ts > current:
            return None
        
        else:
            self._seq = 0

        self._ts = current

        return self._ts << 22 | self._inf | self._seq


import os
from setuptools import (
    setup,
    find_packages
)

current_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(current_directory, 'README.md'), "r") as readme:
    package_description = readme.read()

version_string = ""
with open (os.path.join(current_directory, ".version"), 'r') as version_file:
    version_string = version_file.read()

setup(
    name="mercury-sync",
    version=version_string,
    description="A pure-Python, asyncio based library for authoring distributed systems",
    long_description=package_description,
    long_description_content_type="text/markdown",
    author="Sean Corbett",
    author_email="sean.corbett@umontana.edu",
    url="https://github.com/scorbettUM/mercury-sync",
    packages=find_packages(),
    keywords=[
        'pypi', 
        'python',
        'distributed systems',
        'swim',
        'serf',
        'distributed',
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'pydantic',
        'python3-dtls',
        'zstandard',
        'cryptography',
        'python-dotenv'
    ],
    python_requires='>=3.10'
)



import functools
from mercury_sync.service import Service
from mercury_sync.service.controller import Controller
from typing import Union


def client(
    call_name: str, 
    as_tcp: bool=False
):

    def wraps(func):

        func.client_only = True
        func.target = call_name

        @functools.wraps(func)
        async def decorator(
            *args,
            **kwargs
        ):
            connection: Union[Service, Controller] = args[0]

            if as_tcp:
                return await connection.send_tcp(
                    call_name,
                    await func(*args, **kwargs)
                )

            else:
                return await connection.send(
                    call_name,
                    await func(*args, **kwargs)
                )

        return decorator
    
    return wraps
